/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2015-2016, High Performance Humanoid Technologies (H2T), Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotComponents::ViewSelection
 * @author     David Schiebener (schiebener at kit dot edu)
 * @author     Markus Grotz ( markus dot grotz at kit dot edu )
 * @date       2015
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "ViewSelection.h"

#include <ArmarXCore/core/system/ArmarXDataPath.h>
#include <ArmarXCore/core/system/cmake/CMakePackageFinder.h>

#include <RobotAPI/libraries/core/remoterobot/RemoteRobot.h>


using namespace armarx;


void ViewSelection::onInitComponent()
{
    usingProxy(getProperty<std::string>("RobotStateComponentName").getValue());
    usingProxy(getProperty<std::string>("HeadIKUnitName").getValue());

    offeringTopic("DebugDrawerUpdates");
    offeringTopic(getName() + "Observer");

    headIKKinematicChainName = getProperty<std::string>("HeadIKKinematicChainName").getValue();
    headFrameName = getProperty<std::string>("HeadFrameName").getValue();
    cameraFrameName = getProperty<std::string>("CameraFrameName").getValue();
    drawViewDirection = getProperty<bool>("VisualizeViewDirection").getValue();

    std::string graphFileName = "RobotAPI/ViewSelection/graph40k.gra";

    armarx::CMakePackageFinder finder("RobotAPI");
    ArmarXDataPath::addDataPaths(finder.getDataDir());

    if (ArmarXDataPath::getAbsolutePath(graphFileName, graphFileName))
    {
        visibilityMaskGraph = new CIntensityGraph(graphFileName);
        TNodeList* nodes = visibilityMaskGraph->getNodes();
        const float maxOverallHeadTiltAngle = getProperty<float>("MaxOverallHeadTiltAngle").getValue();
        const float borderAreaAngle = 10.0f;
        const float centralAngle = getProperty<float>("CentralHeadTiltAngle").getValue();

        for (size_t i = 0; i < nodes->size(); i++)
        {
            CIntensityNode* node = (CIntensityNode*) nodes->at(i);

            if (fabs(node->getPosition().fPhi - centralAngle) < maxOverallHeadTiltAngle - borderAreaAngle)
            {
                node->setIntensity(1.0f);
            }
            else if (fabs(node->getPosition().fPhi - centralAngle) < maxOverallHeadTiltAngle)
            {
                node->setIntensity(1.0f - (fabs(node->getPosition().fPhi - centralAngle) - (maxOverallHeadTiltAngle - borderAreaAngle)) / borderAreaAngle);
            }
            else
            {
                node->setIntensity(0.0f);
            }
        }
    }
    else
    {
        ARMARX_ERROR << "Could not find required1 graph file";
        handleExceptions();
        return;
    }

    sleepingTimeBetweenViewDirectionChanges = getProperty<int>("SleepingTimeBetweenViewDirectionChanges").getValue();
    doAutomaticViewSelection = getProperty<bool>("ActiveAtStartup").getValue();

    timeOfLastViewChange =  armarx::TimeUtil::GetTime();

    // this is robot model specific: offset from the used head coordinate system to the actual
    // head center where the eyes are assumed to be located. Here it is correct for ARMAR-III
    offsetToHeadCenter << 0, 0, 150;

    processorTask = new PeriodicTask<ViewSelection>(this, &ViewSelection::process, 30);
    processorTask->setDelayWarningTolerance(sleepingTimeBetweenViewDirectionChanges + 100);
}


void ViewSelection::onConnectComponent()
{
    robotStateComponent = getProxy<RobotStateComponentInterfacePrx>(getProperty<std::string>("RobotStateComponentName").getValue());

    headIKUnitProxy = getProxy<HeadIKUnitInterfacePrx>(getProperty<std::string>("HeadIKUnitName").getValue());
    headIKUnitProxy->request();

    viewSelectionObserver = getTopic<ViewSelectionObserverPrx>(getName() + "Observer");
    drawer = getTopic<DebugDrawerInterfacePrx>("DebugDrawerUpdates");

    processorTask->start();
}


void ViewSelection::onDisconnectComponent()
{
    processorTask->stop();

    if (drawer)
    {
        drawer->removeArrowDebugLayerVisu("HeadRealViewDirection");
    }

    try
    {
        headIKUnitProxy->release();
    }
    catch (...)
    {
        ARMARX_WARNING_S << "An exception occured during onDisconnectComponent()";
        handleExceptions();
    }
}


void ViewSelection::onExitComponent()
{
    delete visibilityMaskGraph;
}


ViewTargetBasePtr ViewSelection::nextAutomaticViewTarget()
{
    IceUtil::Time currentTime =  armarx::TimeUtil::GetTime();
    std::vector<std::string> activeSaliencyMaps;
    for (const auto & p : saliencyMaps)
    {
        if (p.second->validUntil)
        {
            TimestampVariantPtr time = TimestampVariantPtr::dynamicCast(p.second->validUntil);
            if (time->toTime() > currentTime)
            {
                activeSaliencyMaps.push_back(p.second->name);
            }
        }
        else
        {
            activeSaliencyMaps.push_back(p.second->name);
        }

    }

    if (activeSaliencyMaps.empty())
    {
        return nullptr;
    }

    SharedRobotInterfacePrx robotPrx = robotStateComponent->getSynchronizedRobot();
    TNodeList* visibilityMaskGraphNodes = visibilityMaskGraph->getNodes();

    // find the direction with highest saliency
    int maxIndex = -1;
    float maxSaliency = std::numeric_limits<float>::min();
    for (size_t i = 0; i < visibilityMaskGraphNodes->size(); i++)
    {
        float saliency = 0.0f;
        for (const std::string & n : activeSaliencyMaps)
        {
            saliency += saliencyMaps[n]->map[i];
        }

        CIntensityNode* visibilityNode = (CIntensityNode*) visibilityMaskGraphNodes->at(i);
        saliency *= visibilityNode->getIntensity();

        if (saliency > maxSaliency)
        {
            maxSaliency = saliency;
            maxIndex = i;
        }
    }

    ARMARX_DEBUG << "Highest saliency: " << maxSaliency;

    // select a new view if saliency is bigger than threshold (which converges towards 0 over time)
    int timeSinceLastViewChange = (armarx::TimeUtil::GetTime() - timeOfLastViewChange).toMilliSeconds();
    float saliencyThreshold = 0;

    if (timeSinceLastViewChange > 0)
    {
        saliencyThreshold = 2.0f * sleepingTimeBetweenViewDirectionChanges / timeSinceLastViewChange;
    }

    if (maxSaliency > saliencyThreshold)
    {
        Eigen::Vector3f target;
        MathTools::convert(visibilityMaskGraphNodes->at(maxIndex)->getPosition(), target);
        const float distanceFactor = 1500;
        target = distanceFactor * target + offsetToHeadCenter;
        FramedPositionPtr viewTargetPositionPtr = new FramedPosition(target, headFrameName, robotPrx->getName());

        ViewTargetBasePtr viewTarget = new ViewTargetBase();
        viewTarget->priority = armarx::DEFAULT_VIEWTARGET_PRIORITY; // *  maxSaliency;
        viewTarget->timeAdded = new TimestampVariant(armarx::TimeUtil::GetTime());
        viewTarget->target = viewTargetPositionPtr;
        viewTarget->duration = 0;

        return viewTarget;
    }

    return nullptr;
}



void ViewSelection::process()
{
    ViewTargetBasePtr viewTarget;

    if (doAutomaticViewSelection)
    {
        viewTarget = nextAutomaticViewTarget();
    }

    if (!manualViewTargets.empty())
    {
        ScopedLock lock(manualViewTargetsMutex);

        ViewTargetBasePtr manualViewTarget = manualViewTargets.top();

        CompareViewTargets c;

        if (!viewTarget)
        {
            viewTarget = manualViewTarget;
            manualViewTargets.pop();
        }
        else if (c(viewTarget, manualViewTarget))
        {
            viewTarget = manualViewTarget;
            manualViewTargets.pop();
        }
    }

    if (viewTarget)
    {
        SharedRobotInterfacePrx robotPrx = robotStateComponent->getSynchronizedRobot();
        FramedPositionPtr viewTargetPositionPtr = new FramedPosition(FramedPositionPtr::dynamicCast(viewTarget->target)->toEigen(), viewTarget->target->frame, robotPrx->getName());

        if (drawer && robotPrx->hasRobotNode("Cameras") && drawViewDirection)
        {
            float arrowLength = 1500.0f;
            Vector3Ptr startPos = new Vector3(FramedPosePtr::dynamicCast(robotPrx->getRobotNode("Cameras")->getGlobalPose())->toEigen());
            FramedDirectionPtr currentDirection = new FramedDirection(Eigen::Vector3f(1, 0, 0), "Cameras", robotPrx->getName());
            drawer->setArrowDebugLayerVisu("CurrentHeadViewDirection", startPos, currentDirection->toGlobal(robotPrx), DrawColor {0, 0.5, 0.5, 0.1}, arrowLength, 5);
            Eigen::Vector3f plannedDirectionEigen = viewTargetPositionPtr->toGlobalEigen(robotPrx) - startPos->toEigen();
            Vector3Ptr plannedDirection = new Vector3(plannedDirectionEigen);
            drawer->setArrowDebugLayerVisu("PlannedHeadViewDirection", startPos, plannedDirection, DrawColor {0.5, 0.5, 0, 0.1}, arrowLength, 5);
        }

        ARMARX_INFO << "Looking at target " << viewTargetPositionPtr->output();
        headIKUnitProxy->setHeadTarget(headIKKinematicChainName, viewTargetPositionPtr);

        timeOfLastViewChange = armarx::TimeUtil::GetTime();

        float duration = viewTarget->duration;
        if (!viewTarget->duration)
        {
            duration = sleepingTimeBetweenViewDirectionChanges;
        }


        boost::this_thread::sleep(boost::posix_time::milliseconds(duration));
    }
    else
    {
        boost::this_thread::sleep(boost::posix_time::milliseconds(5));
    }
}



void ViewSelection::addManualViewTarget(const FramedPositionBasePtr& target, const Ice::Current& c)
{
    boost::mutex::scoped_lock lock(manualViewTargetsMutex);

    //SharedRobotInterfacePrx sharedRobotInterface = robotStateComponent->getRobotSnapshot();

    ViewTargetBasePtr viewTarget = new ViewTargetBase();
    viewTarget->priority = armarx::DEFAULT_VIEWTARGET_PRIORITY + 1;
    viewTarget->timeAdded = new TimestampVariant(IceUtil::Time::now());
    viewTarget->target = target;
    viewTarget->duration = 0;

    manualViewTargets.push(viewTarget);
}

void ViewSelection::clearManualViewTargets(const Ice::Current& c)
{
    boost::mutex::scoped_lock lock(manualViewTargetsMutex);

    while (!manualViewTargets.empty())
    {
        manualViewTargets.pop();
    }

    // std::priority_queue<ViewTargetBasePtr, vector<ViewTargetBasePtr>, CompareViewTargets> temp;
    // manualViewTargets.swap(temp);
}

ViewTargetList ViewSelection::getManualViewTargets(const Ice::Current& c)
{
    boost::mutex::scoped_lock lock(manualViewTargetsMutex);

    ViewTargetList result;

    std::priority_queue<ViewTargetBasePtr, vector<ViewTargetBasePtr>, CompareViewTargets> temp(manualViewTargets);

    while (!temp.empty())
    {
        result.push_back(temp.top());

        temp.pop();
    }

    return result;
}



void armarx::ViewSelection::updateSaliencyMap(const SaliencyMapBasePtr& map, const Ice::Current&)
{
    boost::mutex::scoped_lock lock(manualViewTargetsMutex);

    saliencyMaps[map->name] = map;
}


PropertyDefinitionsPtr ViewSelection::createPropertyDefinitions()
{
    return PropertyDefinitionsPtr(new ViewSelectionPropertyDefinitions(getConfigIdentifier()));
}



