/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T), Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    ArmarXCore::units
 * @author     Christian Boege (boege at kit dot edu)
 * @date       2011
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */


#include "PlatformUnitSimulation.h"

#include <Eigen/Geometry>
#include <VirtualRobot/MathTools.h>

#include <cmath>

using namespace armarx;

void PlatformUnitSimulation::onInitPlatformUnit()
{
    platformMode = eUndefined;
    referenceFrame = getProperty<std::string>("ReferenceFrame").getValue();
    targetPositionX = currentPositionX = getProperty<float>("InitialPosition.x").getValue();
    targetPositionY = currentPositionY = getProperty<float>("InitialPosition.y").getValue();
    targetRotation = 0.0;
    targetRotation = currentRotation = getProperty<float>("InitialRotation").getValue();

    maxLinearVelocity = getProperty<float>("LinearVelocity").getValue();
    maxAngularVelocity = getProperty<float>("AngularVelocity").getValue();

    positionalAccuracy = 0.01;

    intervalMs = getProperty<int>("IntervalMs").getValue();
    ARMARX_VERBOSE << "Starting platform unit simulation with " << intervalMs << " ms interval";
    simulationTask = new PeriodicTask<PlatformUnitSimulation>(this, &PlatformUnitSimulation::simulationFunction, intervalMs, false, "PlatformUnitSimulation");
}

void PlatformUnitSimulation::onStartPlatformUnit()
{
    listenerPrx->reportPlatformPose(currentPositionX, currentPositionY, currentRotation);
    simulationTask->start();
}

void PlatformUnitSimulation::onStopPlatformUnit()
{
    if (simulationTask)
    {
        simulationTask->stop();
    }
}


void PlatformUnitSimulation::onExitPlatformUnit()
{
    if (simulationTask)
    {
        simulationTask->stop();
    }
}


void PlatformUnitSimulation::simulationFunction()
{
    // the time it took until this method was called again
    double timeDeltaInSeconds = (IceUtil::Time::now() - lastExecutionTime).toMilliSecondsDouble() / 1000.0;
    lastExecutionTime = IceUtil::Time::now();
    std::vector<float> vel(3, 0.0f);
    {
        ScopedLock lock(currentPoseMutex);
        switch (platformMode)
        {
            case ePositionControl:
            {
                float diff = fabs(targetPositionX - currentPositionX);
                ARMARX_INFO << deactivateSpam(0.5) << "Distance to goal X: " << diff;

                if (diff > positionalAccuracy)
                {
                    int sign = copysignf(1.0f, (targetPositionX - currentPositionX));
                    currentPositionX += sign * std::min<float>(maxLinearVelocity * timeDeltaInSeconds, diff);
                    vel[0] = linearVelocityX;
                }

                diff = fabs(targetPositionY - currentPositionY);

                if (diff > positionalAccuracy)
                {
                    int sign = copysignf(1.0f, (targetPositionY - currentPositionY));
                    currentPositionY +=  sign * std::min<float>(maxLinearVelocity * timeDeltaInSeconds, diff);
                    vel[1] = linearVelocityY;
                }

                diff = fabs(targetRotation - currentRotation);

                if (diff > orientationalAccuracy)
                {
                    int sign = copysignf(1.0f, (targetRotation - currentRotation));
                    currentRotation += sign * std::min<float>(maxAngularVelocity * timeDeltaInSeconds, diff);

                    // stay between +/- M_2_PI
                    if (currentRotation > 2 * M_PI)
                    {
                        currentRotation -= 2 * M_PI;
                    }

                    if (currentRotation < -2 * M_PI)
                    {
                        currentRotation += 2 * M_PI;
                    }

                    vel[2] = angularVelocity;

                }
            }
            break;
            case eVelocityControl:
            {
                Eigen::Vector2f targetVel(linearVelocityX, linearVelocityY);
                Eigen::Rotation2Df rot(currentRotation);
                targetVel = rot * targetVel;
                currentPositionX += timeDeltaInSeconds * targetVel[0];
                currentPositionY += timeDeltaInSeconds * targetVel[1];
                currentRotation += angularVelocity * timeDeltaInSeconds;

                // stay between +/- M_2_PI
                if (currentRotation > 2 * M_PI)
                {
                    currentRotation -= 2 * M_PI;
                }

                if (currentRotation < -2 * M_PI)
                {
                    currentRotation += 2 * M_PI;
                }
                vel[0] = linearVelocityY;
                vel[1] = linearVelocityY;
                vel[2] = angularVelocity;
            }
            break;
            default:
                break;
        }
    }
    listenerPrx->reportPlatformPose(currentPositionX, currentPositionY, currentRotation);
    listenerPrx->reportPlatformVelocity(vel[0], vel[1], vel[2]);
}

void PlatformUnitSimulation::moveTo(Ice::Float targetPlatformPositionX, Ice::Float targetPlatformPositionY, Ice::Float targetPlatformRotation, Ice::Float positionalAccuracy, Ice::Float orientationalAccuracy, const Ice::Current& c)
{
    ARMARX_INFO << "new target pose : " << targetPlatformPositionX << ", " << targetPlatformPositionY << " with angle " << targetPlatformRotation;
    {
        ScopedLock lock(currentPoseMutex);
        platformMode = ePositionControl;
        targetPositionX = targetPlatformPositionX;
        targetPositionY = targetPlatformPositionY;
        targetRotation = targetPlatformRotation;
        this->positionalAccuracy = positionalAccuracy;
        this->orientationalAccuracy = orientationalAccuracy;
    }
    listenerPrx->reportNewTargetPose(targetPositionX, targetPositionY, targetRotation);
}


PropertyDefinitionsPtr PlatformUnitSimulation::createPropertyDefinitions()
{
    return PropertyDefinitionsPtr(
               new PlatformUnitSimulationPropertyDefinitions(getConfigIdentifier()));
}


void armarx::PlatformUnitSimulation::move(float targetPlatformVelocityX, float targetPlatformVelocityY, float targetPlatformVelocityRotation, const Ice::Current& c)
{
    ARMARX_INFO << deactivateSpam(1) << "New velocity: " << targetPlatformVelocityX << ", " << targetPlatformVelocityY << " with angular velocity: " << targetPlatformVelocityRotation;
    ScopedLock lock(currentPoseMutex);
    platformMode = eVelocityControl;
    linearVelocityX = std::min(maxLinearVelocity, targetPlatformVelocityX);
    linearVelocityY = std::min(maxLinearVelocity, targetPlatformVelocityY);
    angularVelocity = std::min(maxAngularVelocity, targetPlatformVelocityRotation);

}

void PlatformUnitSimulation::moveRelative(float targetPlatformOffsetX, float targetPlatformOffsetY, float targetPlatformOffsetRotation, float positionalAccuracy, float orientationalAccuracy, const Ice::Current& c)
{
    float targetPositionX, targetPositionY, targetRotation;
    {
        ScopedLock lock(currentPoseMutex);
        targetPositionX = currentPositionX + targetPlatformOffsetX;
        targetPositionY = currentPositionY + targetPlatformOffsetY;
        targetRotation = currentRotation + targetPlatformOffsetRotation;
    }
    moveTo(targetPositionX, targetPositionY, targetRotation, positionalAccuracy, orientationalAccuracy);
}

void PlatformUnitSimulation::setMaxVelocities(float positionalVelocity, float orientaionalVelocity, const Ice::Current& c)
{
    ScopedLock lock(currentPoseMutex);
    maxLinearVelocity = positionalVelocity;
    maxAngularVelocity = orientaionalVelocity;
}

void PlatformUnitSimulation::stopPlatform(const Ice::Current& c)
{
    move(0, 0, 0);
}

